@@ -748,8 +748,7 @@ gnome_rr_screen_finalize (GObject *gobject)
 
     gdk_window_remove_filter (screen->priv->gdk_root, screen_on_event, screen);
 
-    if (screen->priv->info)
-      screen_info_free (screen->priv->info);
+    screen_info_free (screen->priv->info);
 
     G_OBJECT_CLASS (gnome_rr_screen_parent_class)->finalize (gobject);
 }
